/*
 * Copyright (c) 2015 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "user_callbacks.h"

#include <image_util.h>
#include <storage.h>

#define BUFLEN 256

static Evas_Object *image;
static transformation_h handle = NULL;
static media_packet_h media_packet = NULL;
static char *images_directory = NULL;
static char sample_file_path[BUFLEN];
static const char *image_util_source_filename = "sample.jpg";
static char image_util_filename_encoded[BUFLEN];
static const char *image_util_filename_rotate = "rotate.jpg";
static const char *image_util_filename_crop = "crop.jpg";
static const char *image_util_filename_resize = "resize.jpg";
static const char *image_util_filename_color = "colorspace.jpg";
static Evas_Object *btn[4];
static bool transform_finished = false;
static media_packet_h packet_h;

Eina_Bool _btn_enable(void *data)
{
    if (transform_finished) {

        image_util_transform_destroy(handle);
        handle = NULL;

        media_packet_destroy(packet_h);
        packet_h = NULL;

        for (int i = 0; i < 4; i++)
            elm_object_disabled_set(btn[i], EINA_FALSE);

        transform_finished = false;
    }
    return EINA_TRUE;
}

// Callback function invoked when the transformation of the image is finished.
static void _image_util_completed_callback(media_packet_h *dst, int error_code, void *user_data)
{
	packet_h = *dst;
    dlog_print(DLOG_DEBUG, LOG_TAG, "Transformation finished.");
    if (error_code != IMAGE_UTIL_ERROR_NONE || dst == NULL) {
        dlog_print(DLOG_ERROR, LOG_TAG,
                   "An error occurred during transformation.<br>Error code: %d.", error_code);
    } else {
        // Get the transformed image format.
        media_format_h fmt = NULL;

        int error_code = media_packet_get_format(*dst, &fmt);
        if (error_code != MEDIA_PACKET_ERROR_NONE) {
            dlog_print(DLOG_ERROR, LOG_TAG, "media_packet_get_format() failed! Error code = %d",
                       error_code);
            image_util_transform_destroy(handle);
            return;
        }

        // Get the transformed image dimensions and MIME type.
        media_format_mimetype_e mimetype;
        int width, height;

        error_code = media_format_get_video_info(fmt, &mimetype, &width, &height, NULL, NULL);
        if (error_code != MEDIA_FORMAT_ERROR_NONE) {
            dlog_print(DLOG_ERROR, LOG_TAG, "media_format_get_video_info() failed! Error code = %d",
                       error_code);
            media_format_unref(fmt);
            image_util_transform_destroy(handle);
            return;
        }
        // Release the memory allocated for the media format.
        media_format_unref(fmt);

        // Get the buffer where the transformed image is stored.
        void *packet_buffer = NULL;

        error_code = media_packet_get_buffer_data_ptr(*dst, &packet_buffer);
        if (error_code != MEDIA_PACKET_ERROR_NONE) {
            dlog_print(DLOG_ERROR, LOG_TAG,
                       "media_packet_get_buffer_data_ptr() failed! Error code = %d", error_code);
            image_util_transform_destroy(handle);
            return;
        }

        if (mimetype == MEDIA_FORMAT_RGB888 || mimetype == MEDIA_FORMAT_ARGB) {
            // Store the image from the buffer in a file.
            error_code =
                image_util_encode_jpeg(packet_buffer, width, height,
                        (mimetype == MEDIA_FORMAT_RGB888) ? IMAGE_UTIL_COLORSPACE_RGB888 : IMAGE_UTIL_COLORSPACE_ARGB8888,
                                       100, image_util_filename_encoded);
            if (error_code != IMAGE_UTIL_ERROR_NONE) {
                dlog_print(DLOG_ERROR, LOG_TAG, "image_util_encode_jpeg() failed! Error code = %d",
                           error_code);
                image_util_transform_destroy(handle);
                return;
            }

            dlog_print(DLOG_DEBUG, LOG_TAG, "Transformed image file saved at %s",
                       image_util_filename_encoded);

            // Set the path to the image for the EFL image object created earlier, to display the photo.
            elm_image_file_set(image, image_util_filename_encoded, NULL);

            // Check whether an error occurred during loading of the photo.
            Evas_Load_Error error = evas_object_image_load_error_get(image);

            if (EVAS_LOAD_ERROR_NONE != error) {
                dlog_print(DLOG_ERROR, LOG_TAG,
                           "elm_image_file_set() failed! Could not load image: %s. Error message is \"%s\"\n",
                           image_util_filename_encoded, evas_load_error_str(error));
            } else {
                dlog_print(DLOG_DEBUG, LOG_TAG, "elm_image_file_set() successful. Image %s loaded.",
                           image_util_filename_encoded);
                snprintf(sample_file_path, BUFLEN, "%s", image_util_filename_encoded);
                evas_object_show(image);
            }
        } else {
            PRINT_MSG("Can't open the image in unsupported color space.");
            dlog_print(DLOG_DEBUG, LOG_TAG, "Other MIME types are not supported yet.");
        }
    }
    transform_finished = true;
}

static const char *_map_colorspace(image_util_colorspace_e colorspace)
{
    switch (colorspace) {
    case IMAGE_UTIL_COLORSPACE_YV12:
        return "IMAGE_UTIL_COLORSPACE_YV12";

    case IMAGE_UTIL_COLORSPACE_YUV422:
        return "IMAGE_UTIL_COLORSPACE_YUV422";

    case IMAGE_UTIL_COLORSPACE_I420:
        return "IMAGE_UTIL_COLORSPACE_I420";

    case IMAGE_UTIL_COLORSPACE_NV12:
        return "IMAGE_UTIL_COLORSPACE_NV12";

    case IMAGE_UTIL_COLORSPACE_UYVY:
        return "IMAGE_UTIL_COLORSPACE_UYVY";

    case IMAGE_UTIL_COLORSPACE_YUYV:
        return "IMAGE_UTIL_COLORSPACE_YUYV";

    case IMAGE_UTIL_COLORSPACE_RGB565:
        return "IMAGE_UTIL_COLORSPACE_RGB565";

    case IMAGE_UTIL_COLORSPACE_RGB888:
        return "IMAGE_UTIL_COLORSPACE_RGB888";

    case IMAGE_UTIL_COLORSPACE_ARGB8888:
        return "IMAGE_UTIL_COLORSPACE_ARGB8888";

    case IMAGE_UTIL_COLORSPACE_BGRA8888:
        return "IMAGE_UTIL_COLORSPACE_BGRA8888";

    case IMAGE_UTIL_COLORSPACE_RGBA8888:
        return "IMAGE_UTIL_COLORSPACE_RGBA8888";

    case IMAGE_UTIL_COLORSPACE_BGRX8888:
        return "IMAGE_UTIL_COLORSPACE_BGRX8888";

    case IMAGE_UTIL_COLORSPACE_NV21:
        return "IMAGE_UTIL_COLORSPACE_NV21";

    case IMAGE_UTIL_COLORSPACE_NV16:
        return "IMAGE_UTIL_COLORSPACE_NV16";

    case IMAGE_UTIL_COLORSPACE_NV61:
        return "IMAGE_UTIL_COLORSPACE_NV61";
    }
}

// Callback function invoked for every supported color space.
static bool _image_util_supported_jpeg_colorspace_cb(image_util_colorspace_e colorspace,
        void *user_data)
{
    dlog_print(DLOG_DEBUG, LOG_TAG, "%s", _map_colorspace(colorspace));
    PRINT_MSG("%s", _map_colorspace(colorspace));

    // Continue the iteration over all supported color spaces.
    return true;
}

// Callback invoked after clicking any button from the Image Util (except "Clear" button).
static void _image_util_start_cb(void *data, Evas_Object *obj, void *event_info)
{
    transform_finished = false;

    for (int i = 0; i < 4; i++)
        elm_object_disabled_set(btn[i], EINA_TRUE);

    // Decode the given JPEG file to the img_source buffer.
    unsigned char *img_source = NULL;
    int width, height;
    unsigned int size_decode;

    int error_code =
        image_util_decode_jpeg(sample_file_path, IMAGE_UTIL_COLORSPACE_RGB888, &img_source, &width,
                               &height, &size_decode);
    if (error_code != IMAGE_UTIL_ERROR_NONE) {
        PRINT_MSG("image_util_decode_jpeg() failed.");
        dlog_print(DLOG_ERROR, LOG_TAG, "image_util_decode_jpeg() failed! Error code = %d",
                   error_code);
        return;
    }

    dlog_print(DLOG_DEBUG, LOG_TAG, "Decoded image width: %d height: %d size %d", width, height,
               size_decode);

    // Create a media format structure.
    media_format_h fmt;
    error_code = media_format_create(&fmt);
    if (error_code != MEDIA_FORMAT_ERROR_NONE) {
        PRINT_MSG("media_format_create() failed.");
        dlog_print(DLOG_ERROR, LOG_TAG, "media_format_create() failed! Error code = %d",
                   error_code);
        free(img_source);
        return;
    }

    // Set the MIME type of the created format.
    error_code = media_format_set_video_mime(fmt, MEDIA_FORMAT_RGB888);
    if (error_code != MEDIA_FORMAT_ERROR_NONE) {
        PRINT_MSG("media_format_set_video_mime() failed.");
        dlog_print(DLOG_ERROR, LOG_TAG, "media_format_set_video_mime() failed! Error code = %d",
                   error_code);
        media_format_unref(fmt);
        free(img_source);
        return;
    }

    // Set the width of the created format.
    error_code = media_format_set_video_width(fmt, width);
    if (error_code != MEDIA_FORMAT_ERROR_NONE) {
        PRINT_MSG("media_format_set_video_width() failed.");
        dlog_print(DLOG_ERROR, LOG_TAG, "media_format_set_video_width() failed! Error code = %d",
                   error_code);
        media_format_unref(fmt);
        free(img_source);
        return;
    }

    // Set the height of the created format.
    error_code = media_format_set_video_height(fmt, height);
    if (error_code != MEDIA_FORMAT_ERROR_NONE) {
        PRINT_MSG("media_format_set_video_height() failed.");
        dlog_print(DLOG_ERROR, LOG_TAG, "media_format_set_video_height() failed! Error code = %d",
                   error_code);
        media_format_unref(fmt);
        free(img_source);
        return;
    }

    // Create a media packet with the image.
    error_code = media_packet_create_alloc(fmt, NULL, NULL, &media_packet);
    if (error_code != MEDIA_PACKET_ERROR_NONE) {
        PRINT_MSG("media_packet_create_alloc() failed.");
        dlog_print(DLOG_ERROR, LOG_TAG, "media_packet_create_alloc() failed! Error code = %d",
                   error_code);
        media_format_unref(fmt);
        free(img_source);
        return;
    }

    media_format_unref(fmt);

    // Get the pointer to the internal media packet buffer, where the image will be stored.
    void *packet_buffer = NULL;

    error_code = media_packet_get_buffer_data_ptr(media_packet, &packet_buffer);
    if (error_code != MEDIA_PACKET_ERROR_NONE || NULL == packet_buffer) {
        PRINT_MSG("media_packet_get_buffer_data_ptr() failed.");
        dlog_print(DLOG_ERROR, LOG_TAG,
                   "media_packet_get_buffer_data_ptr() failed! Error code = %d", error_code);
        free(img_source);
        return;
    }

    // Copy the image content to the media_packet internal buffer.
    memcpy(packet_buffer, (void *)img_source, size_decode);
    free(img_source);

    // Create a handle to the transformation.
    error_code = image_util_transform_create(&handle);
    if (error_code != IMAGE_UTIL_ERROR_NONE) {
        PRINT_MSG("image_util_transform_create() failed.");
        dlog_print(DLOG_ERROR, LOG_TAG, "image_util_transform_create() failed! Error code = %d",
                   error_code);
        return;
    }

    // Disable the hardware acceleration for the created transformation.
    error_code = image_util_transform_set_hardware_acceleration(handle, false);
    if (error_code != IMAGE_UTIL_ERROR_NONE) {
        dlog_print(DLOG_ERROR, LOG_TAG,
                   "image_util_transform_set_hardware_acceleration() failed! Error code = %d",
                   error_code);
    }

    // Check which option was chosen by the user.
    int selected_operation = (int)data;

    if (selected_operation == 0) {
        PRINT_MSG("<b>Converting the image color space.</b>");
        dlog_print(DLOG_DEBUG, LOG_TAG, "Converting the image color space.");

        // List all available color spaces.

        dlog_print(DLOG_DEBUG, LOG_TAG, "Supported color spaces:");

        error_code = image_util_foreach_supported_jpeg_colorspace(
            _image_util_supported_jpeg_colorspace_cb, NULL);
        if (error_code != IMAGE_UTIL_ERROR_NONE) {
            PRINT_MSG("image_util_foreach_supported_jpeg_colorspace() failed.");
            dlog_print(DLOG_ERROR, LOG_TAG,
                       "image_util_foreach_supported_jpeg_colorspace() failed! Error code = %d",
                       error_code);
            return;
        }

        image_util_colorspace_e colorspace = IMAGE_UTIL_COLORSPACE_ARGB8888;

        // Set the color space the image color space will be converted to.
        error_code = image_util_transform_set_colorspace(handle, colorspace);
        if (error_code != IMAGE_UTIL_ERROR_NONE) {
            PRINT_MSG("image_util_transform_set_colorspace() failed.");
            dlog_print(DLOG_ERROR, LOG_TAG,
                       "image_util_transform_set_colorspace() failed! Error code = %d", error_code);
            return;
        }
        PRINT_MSG("Color space set to %s", _map_colorspace(colorspace));

        snprintf(image_util_filename_encoded, BUFLEN, "%s/%s", images_directory,
                 image_util_filename_color);
    } else if (selected_operation == 1) {
        PRINT_MSG("<b>Halving image dimensions.</b>");
        dlog_print(DLOG_DEBUG, LOG_TAG, "Halving image dimensions.");

        // Set new values for the width and height the image will be resized to.
        unsigned int new_width = ceil(width / 2);
        unsigned int new_height = ceil(height / 2);

        error_code = image_util_transform_set_resolution(handle, new_width, new_height);
        if (error_code != IMAGE_UTIL_ERROR_NONE) {
            PRINT_MSG("image_util_transform_set_resolution() failed.");
            dlog_print(DLOG_ERROR, LOG_TAG,
                       "image_util_transform_set_resolution() failed! Error code = %d", error_code);
            if (width / 2 == 0 || height / 2 == 0) {
                PRINT_MSG("Width = %d, Height = %d. Cannot halve.", width, height);
                PRINT_MSG("Tap 'Clear' button to continue.");
            }
            return;
        }

        PRINT_MSG("New resolution is:%dx%d", new_width, new_height);

        snprintf(image_util_filename_encoded, BUFLEN, "%s/%s", images_directory,
                 image_util_filename_resize);
    } else if (selected_operation == 2) {
        PRINT_MSG("<b>Rotating the image.</b>");
        dlog_print(DLOG_DEBUG, LOG_TAG, "Rotating the image.");

        // Set the rotation value.
        error_code = image_util_transform_set_rotation(handle, IMAGE_UTIL_ROTATION_90);
        if (error_code != IMAGE_UTIL_ERROR_NONE) {
            PRINT_MSG("image_util_transform_set_rotation() failed.");
            dlog_print(DLOG_ERROR, LOG_TAG,
                       "image_util_transform_set_rotation() failed! Error code = %d", error_code);
            return;
        }
        PRINT_MSG("Image rotated by 90 degrees");

        snprintf(image_util_filename_encoded, BUFLEN, "%s/%s", images_directory,
                 image_util_filename_rotate);
    } else {
        PRINT_MSG("<b>Cropping the image.</b>");
        dlog_print(DLOG_DEBUG, LOG_TAG, "Cropping the image.");

        unsigned char *img_source = NULL;
        int width, height;
        unsigned int size_decode;

        int error_code =
            image_util_decode_jpeg(image_util_filename_encoded, IMAGE_UTIL_COLORSPACE_RGB888,
                                   &img_source, &width, &height, &size_decode);
        if (error_code != IMAGE_UTIL_ERROR_NONE) {
            PRINT_MSG("image_util_decode_jpeg().");
            dlog_print(DLOG_ERROR, LOG_TAG, "image_util_decode_jpeg() failed! Error code = %d",
                       error_code);
            return;
        }

        // Set the coordinates that will be used for cropping the image.
        unsigned int start_x, start_y, end_x, end_y;
        start_x = width / 10;
        end_x = (width * 9) / 10;
        start_y = height / 10;
        end_y = (height * 9) / 10;

        if (0 == end_y || 0 == end_x) {
            PRINT_MSG("Unable to crop. Image too small.");
            PRINT_MSG("Tap 'Clear' button to continue.");
            return;
        }

        error_code = image_util_transform_set_crop_area(handle, start_x, start_y, end_x, end_y);
        if (error_code != IMAGE_UTIL_ERROR_NONE) {
            PRINT_MSG("image_util_transform_set_crop_area() failed.");
            dlog_print(DLOG_ERROR, LOG_TAG,
                       "image_util_transform_set_crop_area() failed! Error code = %d", error_code);
            return;
        }
        PRINT_MSG("Cropped area top-left corner:<br>(%d,%d)", start_x, start_y);
        PRINT_MSG("Cropped area bottom-right corner:<br>(%d,%d)", end_x, end_y);
        PRINT_MSG("New resolution: %dx%d", end_x - start_x, end_y - start_y);

        snprintf(image_util_filename_encoded, BUFLEN, "%s/%s", images_directory,
                 image_util_filename_crop);
    }

    // Execute the transformation.
    error_code =
        image_util_transform_run(handle, media_packet, _image_util_completed_callback, NULL);
    if (error_code != IMAGE_UTIL_ERROR_NONE) {
        PRINT_MSG("image_util_transform_run() failed.");
        dlog_print(DLOG_ERROR, LOG_TAG, "image_util_transform_run() failed! Error code = %d",
                   error_code);
    }
}

static bool _storage_cb(int storage_id, storage_type_e type, storage_state_e state,
                        const char *path, void *user_data)
{
    if (STORAGE_TYPE_INTERNAL == type) {
        int *internal_storage_id = (int *)user_data;
        *internal_storage_id = storage_id;

        // Internal storage found, stop the iteration.
        return false;
    } else {
        // Continue iterating over storages.
        return true;
    }
}

// Image Util menu creation.
void create_buttons_in_main_window(appdata_s *ad)
{
    // Create the window for the Image Util.
    Evas_Object *display = _create_new_cd_display(ad, "Image Util", NULL);
    evas_object_size_hint_weight_set(display, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);

    // Create buttons for the Image Util.
    ad->convert_btn = _new_button((void *)0, display, "Convert the Color Space", _image_util_start_cb);
    ad->resize_btn = _new_button((void *)1, display, "Halve the resolution", _image_util_start_cb);
    ad->rotate_btn = _new_button((void *)2, display, "Rotate an Image", _image_util_start_cb);
    ad->crop_btn = _new_button((void *)3, display, "Crop an Image", _image_util_start_cb);

    btn[0] = ad->convert_btn;
    btn[1] = ad->resize_btn;
    btn[2] = ad->rotate_btn;
    btn[3] = ad->crop_btn;

    // Create the top box.
    Evas_Object *top_box = elm_box_add(display);
    evas_object_size_hint_align_set(top_box, EVAS_HINT_FILL, EVAS_HINT_FILL);
    evas_object_size_hint_weight_set(top_box, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
    elm_box_pack_end(display, top_box);
    evas_object_show(top_box);

    // Create table for displaying image on the background.
    Evas_Object *table = elm_table_add(top_box);
    elm_table_homogeneous_set(table, EINA_TRUE);
    evas_object_size_hint_weight_set(table, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
    evas_object_size_hint_align_set(table, EVAS_HINT_FILL, EVAS_HINT_FILL);
    elm_box_pack_end(top_box, table);
    evas_object_show(table);

    // Fill the top box with black background.
    Evas_Object *bg = elm_bg_add(table);
    evas_object_size_hint_align_set(bg, EVAS_HINT_FILL, EVAS_HINT_FILL);
    evas_object_size_hint_weight_set(bg, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
    elm_bg_color_set(bg, 0, 0, 0);
    elm_table_pack(table, bg, 0, 0, 1, 1);
    evas_object_show(bg);

    // Create an image object for displaying the image on it.
    image = elm_image_add(table);
    evas_object_size_hint_align_set(image, EVAS_HINT_FILL, EVAS_HINT_FILL);
    evas_object_size_hint_weight_set(image, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
    elm_table_pack(table, image, 0, 0, 1, 1);
    evas_object_show(image);

    // Get the path to the sample image stored in the resources.
    char *resource_path = app_get_resource_path();
    snprintf(sample_file_path, BUFLEN, "%s%s", resource_path, image_util_source_filename);

    // Set the path to the image file that will be used as a source by the EFL image object.
    elm_image_file_set(image, sample_file_path, NULL);

    // Check whether any error occurred during loading the image.
    Evas_Load_Error error = evas_object_image_load_error_get(image);

    if (EVAS_LOAD_ERROR_NONE != error) {
        dlog_print(DLOG_ERROR, LOG_TAG,
                   "elm_image_file_set() failed! Could not load image: %s. Error message is \"%s\"\n",
                   sample_file_path, evas_load_error_str(error));
    } else {
        dlog_print(DLOG_DEBUG, LOG_TAG, "elm_image_file_set() successful. Image %s loaded.",
                   sample_file_path);
        evas_object_show(image);
    }

    // Get the path to the Images directory:

    // 1. Get internal storage id.
    int internal_storage_id = -1;

    int error_code = storage_foreach_device_supported(_storage_cb, &internal_storage_id);
    if (STORAGE_ERROR_NONE != error_code) {
        dlog_print(DLOG_ERROR, LOG_TAG,
                   "storage_foreach_device_supported() failed! Error code = %d", error_code);
        free(resource_path);
        return;
    }

    // 2. Get the path to the Images directory.
    error_code =
        storage_get_directory(internal_storage_id, STORAGE_DIRECTORY_IMAGES, &images_directory);
    if (STORAGE_ERROR_NONE != error_code) {
        dlog_print(DLOG_ERROR, LOG_TAG, "storage_get_directory() failed with error code: %d",
                   error_code);
    }

    snprintf(image_util_filename_encoded, BUFLEN, "%s/%s", resource_path,
             image_util_source_filename);
    free(resource_path);

    ecore_idler_add(_btn_enable, NULL);
}

// Callback function invoked when the "Clear" button is clicked in the Image Util.
void _image_util_clear_cb(void *data, Evas_Object *obj, void *event_info)
{
    // Get the path to the sample image stored in the resources.
    char *resource_path = app_get_resource_path();
    snprintf(sample_file_path, BUFLEN, "%s/%s", resource_path, image_util_source_filename);
    snprintf(image_util_filename_encoded, BUFLEN, "%s/%s", resource_path,
             image_util_source_filename);
    free(resource_path);

    // Set the path to the image file that will be used as a source by the EFL image object.
    elm_image_file_set(image, sample_file_path, NULL);

    // Check whether any error occurred during loading the photo.
    Evas_Load_Error error = evas_object_image_load_error_get(image);

    if (EVAS_LOAD_ERROR_NONE != error) {
        dlog_print(DLOG_ERROR, LOG_TAG,
                   "elm_image_file_set() failed! Could not load image: %s. Error message is \"%s\"\n",
                   sample_file_path, evas_load_error_str(error));
    } else {
        dlog_print(DLOG_DEBUG, LOG_TAG, "elm_image_file_set() successful. Image %s loaded.",
                   sample_file_path);
        evas_object_show(image);
    }

    for (int i = 0; i < 4; i++)
        elm_object_disabled_set(btn[i], EINA_FALSE);
}
